char
    VERSION1 = '2',
    VERSION2 = '3';

bool
    SMALL = false,
    MEDIUM = false,
    SPECIAL = false,
    GOTIO = true ,		/* true -> read/write constructs exist */
    GOTPEEP = true ,		/* true -> peephole optimizer exists */
    GOTFLOAT = true ,		/* true -> 32 bit & floating point exist */
    GOTCOPY = true ;		/* true -> copyright protection installed */

word
    DRACOMAGIC = 0x4688,

    FRACLEN = 7,	/* number of bytes in floating point fraction */

    PBSIZE = if SPECIAL then 8000 elif SMALL then 4000 else 6000 fi,
			/* size of code/symbol buffer */
    SYSIZE = if SMALL then 512 else 701 fi,
			/* size of symbol table */
    CBSIZE = if SMALL then 300 elif MEDIUM then 300 else 500 fi,
			/* size of constant buffer */
    CTSIZE = if SMALL then 20 elif MEDIUM then 20 else 35 fi,
			/* size of constant table */
    DTSIZE = 10,	/* size of descriptor table */
    GRTSIZE = 50,	/* size of globals relocation table */
    FRTSIZE = if SMALL then 50 elif MEDIUM then 50 else 75 fi,
			/* size of file statics relocation table */
    LRTSIZE = 50,	/* size of static locals relocation table */
    PRTSIZE = if SMALL then 100 elif MEDIUM then 100 else 200 fi,
			/* size of program label relocation table */
    CASESIZE = if SMALL then 75 elif MEDIUM then 75 else 100 fi,
			/* size of case alternative table */
    TTSIZE = if SMALL then 175 else 200 fi,
			/* size of type table (< 256) */
    TITSIZE = if SMALL then 500 else 600 fi,
			/* size of type info table */
    SECSIZE = 128;	/* size of CP/M sector */

type

    /* codes for kinds of types: */

    TYPEKIND = enum {
	TY_POINTER,
	TY_ENUM,
	TY_UNSIGNED,
	TY_SIGNED,
	TY_FILE,
	TY_CHANNEL,
	TY_PRIVATE,
	TY_STRUCT,
	TY_ARRAY,
	TY_OP,
	TY_PROC,
	TY_SPECIAL,
	TY_UNDEFINED,
	TY_NAMED
    },

    /* flags for the two types of array dimension entries: */

    ARRAYKIND = enum {
	AR_FIXED,
	AR_FLEX
    },

    /* the type of a type, as it is used in descriptors: */

    TYPENUMBER = unsigned TTSIZE,

    /* the structure of an entry in the symbol table: */

    SYMBOL = struct {
	ushort sy_kind;
	TYPENUMBER sy_type;
	*char sy_name;
	word sy_value;
    },

    /* the structure of an array dimension in the type info data: */

    ARRAYDIM = struct {
	ARRAYKIND ar_kind;
	word ar_dim;
    },

    /* the structure of an array in the type info data: */

    ARRAYDESC = struct {
	TYPENUMBER ar_baseType;
	ushort ar_dimCount;
	[1] ARRAYDIM ar_dims;
    },

    /* the structure of a struct/union in the type info data: */

    STRUCTDESC = struct {
	word st_size;
	ushort st_fieldCount;
	[1] *SYMBOL st_fields;
    },

    /* the structure of a proc type in the type info data: */

    PROCDESC = struct {
	TYPENUMBER p_resultType;
	ushort p_parCount;
	[1] TYPENUMBER p_parTypes;
    },

    /* the structure of an operator type in the type info data: */

    OPDESC = struct {
	TYPENUMBER op_baseType;
	word op_ops;
	[2] char op_name;
    },

    /* the united type for a type info value: */

    INFOTYPE = union {
	word i_word;
	TYPENUMBER i_type;
	*byte i_ptr;
	*ARRAYDESC i_array;
	*STRUCTDESC i_struct;
	*PROCDESC i_proc;
	*OPDESC i_op;
	struct {
	    bool i_input;
	    bool i_text;
	} i_channel;
    },

    /* the structure of an entry in the type table: */

    TTENTRY = struct {
	TYPEKIND t_kind;
	INFOTYPE t_info;
    },

    /* the nature of a value in a descriptor: */

    VALUEKIND = enum {
	VERROR,
	VVOID,
	VPROC,
	VAPROC,
	VNUMBER,
	VCONST,
	VGVAR,
	VFVAR,
	VLVAR,
	VAVAR,
	VPAR,
	VDEREF,
	VEXPR,
	VCC
    },

    /* the possible values for a comparison result: */

    COMPARISONKIND = enum {
	VEQ,
	VNE,
	VGT,
	VLT,
	VGE,
	VLE,
	VGTS,
	VLTS,
	VGES,
	VLES
    },

    /* the structure of an entry in the constant table: */

    CTENT = struct {
	byte ct_type;
	*word ct_use;
	*byte ct_value;
	word ct_length;
    },

    /* the united type for a descriptor value field: */

    VALUETYPE = union {
	COMPARISONKIND v_comparison;
	word v_word;
	int v_int;
	byte v_byte;
	*word v_proc;
	*CTENT v_const;
    },

    /* the structure of a value descriptor: */

    DESCRIPTOR = struct {
	VALUEKIND v_kind;
	TYPENUMBER v_type;
	VALUETYPE v_value;
    },

    /* the structure of a relocation table entry: */

    RELOC = struct {
	word r_what;
	word r_head;
    },

    /* the structure of an entry in the case alternative table: */

    CENTRY = struct {
	word c_index;
	*byte c_codePointer;
    },

    /* the structure of a CP/M file control block: */

    FCB = struct {
	byte fcb_drive;
	[8] char fcb_name;
	[3] char fcb_type;
	byte fcb_extent;
	byte fcb_s1;
	byte fcb_s2;
	byte fcb_recordCount;
	[16] byte fcb_alternate;
	byte fcb_currentRecord;
	word fcb_position;
	byte fcb_positionHigh;
    };

/* flag bits for the operators defined for an operator type: */

word
    OPADD = 0b0000000000000001,
    OPSUB = 0b0000000000000010,
    OPMUL = 0b0000000000000100,
    OPDIV = 0b0000000000001000,
    OPMOD = 0b0000000000010000,
    OPNEG = 0b0000000000100000,
    OPABS = 0b0000000001000000,
    OPIOR = 0b0000000010000000,
    OPAND = 0b0000000100000000,
    OPXOR = 0b0000001000000000,
    OPSHL = 0b0000010000000000,
    OPSHR = 0b0000100000000000,
    OPNOT = 0b0001000000000000,
    OPCPR = 0b0010000000000000,
    OPPUT = 0b0100000000000000,
    OPGET = 0b1000000000000000;

/* the special types known to the compiler: */

TYPENUMBER
    TYUNKNOWN = 1,
    TYERROR = 2,
    TYIORESULT = 4,
    TYVOID = 5,
    TYNIL = 6,
    TYBOOL = 8,
    TYCHAR = 10,
    TYBYTE = 12,
    TYUSHORT = 14,
    TYSHORT = 16,
    TYWORD = 18,
    TYINT = 20,
    TYCHARS = 21,
    TYLONG = 24,
    TYULONG = 27,
    TYFLOAT = 30;

byte

/* codes for the types of constants in the constant table: */

    C_TABLE = 0x00,
    C_INLINE = 0x01,

/* codes for the 'kind' field in symbol table entries: */
	
/* the top two bits are used for the level of declaration: */
	
    BB = 0xc0,

    B_SYS = 0x00,			/* system symbol, e.g. 'int' */
    B_GLOBAL = 0x40,			/* global from include file */
    B_FILE = 0x80,			/* file global symbol */
    B_LOCAL = 0xc0,			/* local to a given proc */

/* the rest of the bits are used for the nature of the symbol: */
	
    MMMMMM = 0x3f,

    MFREE = 0,
    MUNDEF = 1,
    MKEYW = 2,
    MPROC = 3,
    MAPROC = 4,
    MNUMBER = 5,
    MCONST = 6,
    MGVAR = 7,
    MFVAR = 8,
    MLVAR = 9,
    MAVAR = 10,
    MPAR = 11,
    MFIELD = 12,
    MTYPE = 13;

/* the various special values for tokens: */

ushort
    TEOF	= 0,
    TOKERR	= TEOF		+ 1,
    TNUMBER	= TOKERR	+ 1,
    TCHAR	= TNUMBER	+ 1,
    TCHARS	= TCHAR		+ 1,
    TLNUM	= TCHARS	+ 1,
    TFNUM	= TLNUM		+ 1,
    TID		= TFNUM		+ 1,

    TNE		= TID		+ 1,
    TLE		= TNE		+ 1,
    TGE		= TLE		+ 1,
    TASS	= TGE		+ 1,
    TSHL	= TASS		+ 1,
    TSHR	= TSHL		+ 1,
    TXOR	= TSHR		+ 1,
    TDOTDOT	= TXOR		+ 1,

    TSIGNED	= 128,
    TUNSIGNED	= TSIGNED	+ 1,
    TENUM	= TUNSIGNED	+ 1,
    TSTRUCT	= TENUM		+ 1,
    TUNION	= TSTRUCT	+ 1,
    TPRIVATE	= TUNION	+ 1,
    TFILE	= TPRIVATE	+ 1,
    TCHANNEL	= TFILE		+ 1,

    TTEXT	= TCHANNEL	+ 1,
    TBINARY	= TTEXT		+ 1,
    TTYPE	= TBINARY	+ 1,
    TEXTERN	= TTYPE		+ 1,
    TNONREC	= TEXTERN	+ 1,
    TVECTOR	= TNONREC	+ 1,
    TAND	= TVECTOR	+ 1,
    TOR 	= TAND		+ 1,
    TTHEN	= TOR		+ 1,
    TFROM	= TTHEN		+ 1,
    TUPTO	= TFROM		+ 1,
    TDOWNTO	= TUPTO		+ 1,
    TBY 	= TDOWNTO	+ 1,

    TNIL	= TBY		+ 1,
    TDIM	= TNIL		+ 1,
    TSIZEOF	= TDIM		+ 1,
    TRANGE	= TSIZEOF	+ 1,
    TMAKE	= TRANGE	+ 1,
    TINPUT	= TMAKE		+ 1,
    TNEW	= TINPUT	+ 1,
    TIOERROR	= TNEW		+ 1,
    TNOT	= TIOERROR	+ 1,

    TIF 	= TNOT		+ 1,
    TCASE	= TIF		+ 1,
    TPRETEND	= TCASE		+ 1,
    TOPEN	= TPRETEND	+ 1,
    TCLOSE	= TOPEN		+ 1,
    TREAD	= TCLOSE	+ 1,
    TREADLN	= TREAD		+ 1,
    TWRITE	= TREADLN	+ 1,
    TWRITELN	= TWRITE	+ 1,

    TPROC	= TWRITELN	+ 1,
    TCORP	= TPROC		+ 1,
    TELIF	= TCORP		+ 1,
    TELSE	= TELIF		+ 1,
    TFI 	= TELSE		+ 1,
    TDO 	= TFI		+ 1,
    TOD 	= TDO		+ 1,
    TINCASE	= TOD		+ 1,
    TDEFAULT	= TINCASE	+ 1,
    TESAC	= TDEFAULT	+ 1,

    TWHILE	= TESAC		+ 1,
    TOUTPUT	= TWHILE	+ 1,
    THALT	= TOUTPUT	+ 1,
    TION	= THALT		+ 1,
    TIOFF	= TION		+ 1,
    TFOR	= TIOFF		+ 1,
    TFREE	= TFOR		+ 1,
    TCODE	= TFREE		+ 1,
    TERROR	= TCODE		+ 1;

byte

    TMASK = 0x96,

/* hardware specifics: machine registers and op-codes: */

    RA = 7,
    RB = 0,
    RC = 1,
    RD = 2,
    RE = 3,
    RH = 4,
    RL = 5,
    RM = 6,
    RSP = 6,
    RPSW = 6,

    ONOP = 0o000,
    OJMP = 0o303,
    OJNZ = 0o302,
    OJZ = 0o312,
    OJNC = 0o322,
    OJC = 0o332,
    OJP = 0o362,
    OJM = 0o372,
    OMOV = 0o100,
    OHLT = 0o166,
    OMVI = 0o006,
    OINR = 0o004,
    ODCR = 0o005,
    OADD = 0o200,
    OSUB = 0o220,
    OANA = 0o240,
    OXRA = 0o250,
    OORA = 0o260,
    OCMP = 0o270,
    OADI = 0o306,
    OSUI = 0o326,
    OANI = 0o346,
    OXRI = 0o356,
    OORI = 0o366,
    OCPI = 0o376,
    ORLC = 0o007,
    ORRC = 0o017,
    ORAL = 0o027,
    ORAR = 0o037,
    OCALL = 0o315,
    OIN = 0o333,
    OOUT = 0o323,
    OEI = 0o373,
    ODI = 0o363,
    OLXI = 0o001,
    OPUSH = 0o305,
    OPOP = 0o301,
    OLDA = 0o072,
    OSTA = 0o062,
    ODAD = 0o011,
    OINX = 0o003,
    ODCX = 0o013,
    OCMA = 0o057,
    OSHLD = 0o042,
    OLHLD = 0o052,
    OPCHL = 0o351,
    ORET = 0o311,
    OXCHG = 0o353;

FCB
    CodeFile,				/* output to .REL file */
    MainFile,				/* input from .DRC file */
    IncludeFile,			/* input from .G file */
    ErrorFile;				/* error message database */
*FCB
    SourceFile;				/* pointer to current input FCB */

[SECSIZE] char
    MainBuff,				/* input buffer for normal mode */
    IncludeBuff;			/* input buffer during globals */
*[SECSIZE] char SourceBuff;		/* pointer to current input buffer */
[SECSIZE] byte CodeBuff;		/* buffer for code output */

*char
    Trap;				/* guess what this is for! */

ushort
    SourcePos,				/* pos in current input buffer */
    CodePos,				/* current position in code buffer */
    PeepNext;				/* number of entries in peephole */

word
    Line,				/* current Line in input file */
    OLine,				/* line for previous char */
    OOLine,				/* line for 2nd previous char */
    OldLine,				/* Line for previous token */
    ErrorCount,				/* number of errors encountered */
    IntValue,				/* numeric value from scanner */
    DeclOffset,				/* offset during declarations */
    PeepTotal,				/* number of instructions gen'd */
    ConstLength;			/* length so far of constant */

int
    HLOffset,				/* current offset of HL from var */
    AHLOffset;				/* alternate offset */

/*
[PBSIZE] byte ProgramBuff @ 0xe000;
*/
[PBSIZE] byte ProgramBuff;		/* program and symbol table buffer */

[PBSIZE / sizeof(word)] word ProgramBuffWord @ ProgramBuff;
*byte ProgramNext;			/* code pointer */
*char
    SymNext;				/* next char for symbol names */
*word
    ProgramNextWord @ ProgramNext,	/* code pointer for chaining */
    TrueChain,				/* true branch chain */
    FalseChain;				/* false branch chain */

[SYSIZE] SYMBOL SymbolTable;		/* the symbol table */
SYMBOL DummyId;				/* dummy symbol entry */
*SYMBOL CurrentId @ IntValue;		/* currently scanned identifier */

[CTSIZE] CTENT ConstTable;		/* constant table (CTENT's) */
*CTENT
    ConstNext,				/* pointer to next constant slot */
    String @ IntValue;			/* string returned from scanner */

[DTSIZE] DESCRIPTOR DescTable;		/* descriptor table (DESCRIPTOR's) */
*DESCRIPTOR DescNext;			/* pointer to next descriptor slot */

DESCRIPTOR
    CurrentA,				/* value currently in A reg. */
    CurrentHL,				/* value currently in HL pair */
    AlternateA,				/* other value now in A reg. */
    AlternateHL,			/* other value now in HL pair */
    LoadedValue;			/* last value loaded with LHLD */

[GRTSIZE] RELOC GlobalRelocTable;	/* globals relocation table*/
[FRTSIZE] RELOC FileRelocTable;		/* file statics reloc table*/
[LRTSIZE] RELOC LocalRelocTable;	/* locals relocation table */
[PRTSIZE] RELOC ProgramRelocTable;	/* prog labels reloc table */
*RELOC
    GlobalRelocNext,			/* pointer to next gr slot */
    FileRelocNext,			/* pointer to next fr slot */
    LocalRelocNext,			/* pointer to next lr slot */
    ProgramRelocNext;			/* pointer to next pr slot */

[CBSIZE] byte ByteBuff;			/* constant byte buffer */
*byte ByteNext;				/* pointer into byte buffer */

[CASESIZE] CENTRY CaseTable;		/* table for cases */
*CENTRY CaseTableNext;			/* next slot in case table */

[TTSIZE] TTENTRY TypeTable;		/* the type table */
TYPENUMBER NextType;			/* next available type slot */

[TITSIZE] byte TypeInfoTable;		/* type information table */
*byte NextTypeInfo;			/* next slot in type info table */

char
    Char,				/* current input character */
    NextChar;				/* next input character */

byte
    DeclLevel;				/* current declaration level */

ushort
    Token,				/* code for current token */
    Eof,				/* true if have hit eof */
    Column,				/* current column in source file */
    OColumn,				/* column for previous char */
    OOColumn,				/* column for 2nd previous char */
    OldColumn;				/* column for previous token */

bool
    Ignore,				/* true if ignoring code */
    ABusy,				/* true if reg A is currently busy */
    HLBusy,				/* true if HL pair currently busy */
    ErrorOpen,				/* true if error file opened */
    NamePrinted,			/* true if file had error */
    InLine,				/* true if inside a proc now */
    HLAddress,				/* true when HL has address of var */
    InConst;				/* true when building constant */

/*
 * variables mapped onto by various low-level routines:
 */

word W1, W2, W3, W4, W5;
byte B1, B2, B3, B4, B5, B6, B7;
